

module AES_Decrypt(data_set,clock,reset,key,decrypt_data);


input [7:0] data_set [47:0];  //3 SETS OF 16 BIT DATA
input clock;
input reset;
input [7:0] key [15:0];   //KEY
output [7:0] decrypt_data [47:0];  //DECRYPTED DATA

reg key_gen_done;
reg [7:0] keys_out    [9:0][3:0][3:0];
reg [7:0] addr [3:0];
reg [7:0] data [3:0];
reg [7:0] key_matrix    [3:0][3:0];

//MATRICES TO STORE THE INPUT
reg [7:0] state_matrix0 [3:0][3:0];
reg [7:0] state_matrix1 [3:0][3:0];

//REGISTERS TO STORE OUTPUT FORM EVERY ROUND
reg [7:0] d0_out        [3:0][3:0]; 
reg [7:0] d1_out        [3:0][3:0]; 
reg [7:0] d2_out        [3:0][3:0]; 
reg [7:0] d3_out        [3:0][3:0]; 
reg [7:0] d4_out        [3:0][3:0]; 
reg [7:0] d5_out        [3:0][3:0]; 
reg [7:0] d6_out        [3:0][3:0]; 
reg [7:0] d7_out        [3:0][3:0];
reg [7:0] d8_out        [3:0][3:0]; 
reg [7:0] d9_out        [3:0][3:0];  

reg d_done[9:0];
reg enable_4;

reg new_word;

reg [7:0] decrypt_out [47:0];

reg enable;
reg [5:0] state;
reg [4:0] i,j,k,n,n1;
parameter key_expansion = 4'd0,get_data=4'd1, init_add_round_key = 4'd2, wait_state = 4'd3;

//KEY EXPANSION MODULE			 
key_generator K(
						.clk(clock),
						.reset(reset),
						.initial_key(key_matrix),
						.keys(keys_out),
						.key_gen_done(key_gen_done)
					);

	
//DECRYPTOR MODULES	
 decrypter_pipe d0(   
                    .data(state_matrix1),
                    .clock(clock),
						  .reset(reset | ~key_gen_done | new_word ),
						  .expanded_key(keys_out[8]),
						  .round_num(4'd1),
						  .out_data(d0_out),
						  .round_over(d_done[0])

                  ); 
 
 decrypter_pipe d1(   
                    .data(d0_out),
                    .clock(clock),
						  .reset(d_done[0]),
						  .expanded_key(keys_out[7]),
						  .round_num(4'd2),
						  .out_data(d1_out),
						  .round_over(d_done[1])

                  );
					
 decrypter_pipe d2(   
                    .data(d1_out),
                    .clock(clock),
						  .reset(d_done[1]),
						  .expanded_key(keys_out[6]),
						  .round_num(4'd3),
						  .out_data(d2_out),
						  .round_over(d_done[2])

                  ); 
							
 decrypter_pipe d3(   
                    .data(d2_out),
                    .clock(clock),
						  .reset(d_done[2]),
						  .expanded_key(keys_out[5]),
						  .round_num(4'd4),
						  .out_data(d3_out),
						  .round_over(d_done[3])

                  ); 
						
 decrypter_pipe d4(   
                    .data(d3_out),
                    .clock(clock),
						  .reset(d_done[3]),
						  .expanded_key(keys_out[4]),
						  .round_num(4'd5),
						  .out_data(d4_out),
						  .round_over(d_done[4])

                  ); 
						

						

						
 decrypter_pipe d5(   
                    .data(d4_out),
                    .clock(clock),
						  .reset(d_done[4]),
						  .expanded_key(keys_out[3]),
						  .round_num(4'd6),
						  .out_data(d5_out),
						  .round_over(d_done[5])

                  ); 
						
 decrypter_pipe d6(   
                    .data(d5_out),
                    .clock(clock),
						  .reset(d_done[5]),
						  .expanded_key(keys_out[2]),
						  .round_num(4'd7),
						  .out_data(d6_out),
						  .round_over(d_done[6])

                  ); 
						
						
 decrypter_pipe d7(   
                    .data(d6_out),
                    .clock(clock),
						  .reset(d_done[6]),
						  .expanded_key(keys_out[1]),
						  .round_num(4'd8),
						  .out_data(d7_out),
						  .round_over(d_done[7])

                  ); 
						
 decrypter_pipe d8(   
                    .data(d7_out),
                    .clock(clock),
						  .reset(d_done[7]),
						  .expanded_key(keys_out[0]),
						  .round_num(4'd9),
						  .out_data(d8_out),
						  .round_over(d_done[8])

                  ); 
						
 decrypter_pipe d9(   
                    .data(d8_out),
                    .clock(clock),
						  .reset(d_done[8]),
						  .expanded_key(key_matrix),
						  .round_num(4'd10),
						  .out_data(d9_out),
						  .round_over(d_done[9])

                  ); 
						
						
						
						
						
						
						
						
						
						
						

always@(posedge clock) begin
if(reset) begin
	enable<=0;
	state<= key_expansion;
		for(i=0;i<4;i=i+1) begin
		   for(j=0;j<4;j=j+1)begin
			key_matrix[j][i] <= key[i*4 + j];

		  end
	   end

	n <= 0;
	n1 <= 0;
	new_word <= 0;
end
else begin
   case(state)
	//WAIT FOR COMPLETION OF KEY EXPANSION
	 key_expansion: begin
	   if(key_gen_done == 1) begin
	     state <= get_data;
		end
	  end
	 //AFTER KEY EXPANSION, GET THE INPUT
	  get_data: begin
			for(i=0;i<4;i=i+1) begin
				for(j=0;j<4;j=j+1) begin
					state_matrix0[j][i] <= data_set[n*16 + i*4 + j];
				end
			end
			state <= init_add_round_key;
		end
	  //PERFORM INITIAL XOR OPERATION WITHT HE KEY
	  init_add_round_key: begin
			for(i=0;i<4;i=i+1) begin
				for(j=0;j<4;j=j+1) begin
					state_matrix1[j][i] <= state_matrix0[j][i] ^ keys_out[9][j][i];
				end
			end
		state <= wait_state;
		new_word <= 0;
		end
		//IF LAST ROUND COMPLETED TRANSFER THE RESULTS TO OUTPUT
		wait_state: begin
			if(d_done[9] == 1) begin
				if(n1 >0) begin
				for(i=0;i<4;i=i+1) begin
					for(j=0;j<4;j=j+1) begin
						decrypt_out[(n1-1)*16 + i*4 + j] <= d9_out[j][i];
					end
				end
				end
				n1 <= n1+1;	
			end
			//WAIT FOR THE FIRST ROUND TO GET OVER IN ORDER TO PROCESS THE NEXT SET OF CHARACTERS
			if(d_done[0] == 1) begin
				if(n1 <3) begin
				new_word <= 1;
				state <= get_data;
				n <= n+1;
				end
			end
		end
						
	endcase
end
end

assign decrypt_data = decrypt_out;

endmodule