/*
ECE 4760 (Fall 2014)
Andrew Wilson, Ashkan Ravani, Divyanshu Tiwari

<<<DESCRIPTION, INTRO GOES HERE>>>
*/

#define F_CPU 16000000UL
#include <avr/io.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <avr/eeprom.h>
#include <math.h> 			//For functions sqrt() and atan()

#define OVERHEAD 250.0			//Experimentally estimated total calculation time in process
#define x1 27.7				//Half the cop car width, in units of 62.5usec*343 m/s
#define PI 3.14159

//Coordinate variables
volatile unsigned int tA, tB;		//Overall path times for A, B
float x, y, theta;
int angle;

//PID variables
float prev_angle=0, errorD_A=0, errorI_A=0;
float pGain_A, dGain_A, iGain_A;
float output_A=0;
int duty_A=0;
int temp=0;

int distance_setPoint=75;
float error;
float pGain_D, dGain_D, iGain_D;
float prev_dist=0, errorD_D=0, errorI_D=0;
float output_D=0;
int neg=0; //indicates if angle is negative.  used in PID control
int duty_D=0;
int pulseLength=50; //no more than ~6 mSec, probably about 1 mSec
int pulse_start;
volatile int noDataForPulses=0;	//Record the number of pulse periods for which new coordinate data hasn't been computed

volatile char int2executed = 0;	//State variable - has INT2 ISR executed for the current pulse?
volatile char int1executed = 0;	//State variable - has INT1 ISR executed for the current pulse?

//Interrupt Service Routine for rising edge detection on port B2,
//which is attached to Receiver A at coordinate (-x1,0) i.e., the
//receiver on the right.
ISR(INT2_vect){

	tA = TCNT3;
	int2executed = 1;
	EIMSK= 2;		//Disable INT2, INT1 remains enabled until next pulse
	EIFR=1<<INTF2;
	if(int1executed){
		EIMSK= 0;
		EIFR=((1<<INTF2)|(1<<INTF1));
		noDataForPulses = 0;	//Coordinate computation occurs
		coordinates();	//On the final interrupt, execute coordinate computation
		PIDcontrol();	//then run PID control to operate motor
		
	}
} 

//Interrupt Service Routine for rising edge detection on port D3,
//which is attached to Receiver B at coordinate (+x1,0) i.e., the
//receiver on the left.
ISR(INT1_vect){
	tB = TCNT3;
	int1executed = 1;
	EIMSK= 4;		//Disable INT1, INT2 remains enabled until next pulse
	EIFR=1<<INTF1;
	if(int2executed){
		EIMSK= 0;
		EIFR=((1<<INTF2)|(1<<INTF1));
		noDataForPulses = 0;	//Coordinate computation occurs
		coordinates();	//On the final interrupt, execute coordinate computation
		PIDcontrol();	//then run PID control to operate motor
		
	}
}


//////////////////////////////////////////////////////////////////
ISR (TIMER3_COMPA_vect){

TCNT3=0;
int2executed = 0;
int1executed = 0;
noDataForPulses++;
if(noDataForPulses==10){	//PID control has likely overshot and lost signal data
	Stopped();
}
emitPulse();
//Turn interrupts back on
EIMSK = 6;
EIFR=((1<<INTF2)|(1<<INTF1));

}
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
void turn_Right(void){

TCCR0A = (0<<COM0A0) | (1<<COM0A1) |(0<<COM0B0) | (0<<COM0B1)| (1<<WGM00) | (1<<WGM01) ;

}
//////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////
void turn_Left(void){

TCCR0A = (0<<COM0A0) | (0<<COM0A1) |(0<<COM0B0) | (1<<COM0B1)| (1<<WGM00) | (1<<WGM01) ;

}
///////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////
void no_turn(void){

TCCR0A = (0<<COM0A0) | (0<<COM0A1) |(0<<COM0B0) | (0<<COM0B1)| (1<<WGM00) | (1<<WGM01) ;

}
////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////
void Forward(void){

TCCR2A = (0<<COM2A0) | (1<<COM2A1) |(0<<COM2B0) | (0<<COM2B1)| (1<<WGM20) | (1<<WGM21) ;

}
////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////
void Reverse(void){

TCCR2A = (0<<COM2A0) | (0<<COM2A1) |(0<<COM2B0) | (1<<COM2B1)| (1<<WGM20) | (1<<WGM21) ;

}
////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////
void Stopped(void){

OCR2A=0;

}
////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
//Perform coordinate calculation
//Since the distance magnitude is irrelevant to us, we assume a scaling in which
//sound waves have unit velocity. Thus, distance = time.
void coordinates(){
//Account for path doubling and microcontroller overhead
tA = (tA - OVERHEAD)/2.0;			//A has transmitter and receiver
tB = tB - (tA + OVERHEAD);		//tA is now outgoing path duration/length
//tA, tB & tC are now proportional to the distances of the target from A, B and C respectively
//Now we compute theta and y. See project documentation for underlying mathematics

///NOTE: We assume the mobile target is in the half-plane in front of the chasing unit.
///This gives us a unique position using two receivers on the chasing unit.
///In theory, the third receiver could give an exact position (x,y) on the entire plane,
///but the directionality of the speakers and microphones makes that practically unfeasible
///With this assumption, we only have a transmitter and a receiver at A(-x1,0).

//Calculate the angular position of the chase target with respect to the axis in
//parallel to the direction of the chasing unit (y-axis in our coordinate system)
	if(tA>tB){
		theta = asin((tA-tB)/(2.0*x1));
	} else{
		theta = -1*asin((tB-tA)/(2.0*x1));
	}

	y = cos(theta)*(tA+tB)/2;	//Distance of target system along y-axis

//Positive angle is in counter-clockwise direction
//For example, positive x-axis (right) is at -pi/2, positive y-axis at 0,
//and negative x-axis (left) at pi/2
	angle = (theta)*(180/PI);	//Get angle in degrees

}
//////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////
//A PID control loop system to operate the motor control
//The following parameters have been used:
//Proportional gain = ; Integral gain = ; Derivative gain = ;
//The error is the value of the deviation angle theta, in radians
//If theta < 0, the target is to the right of the chasing unit's current orientation
void PIDcontrol(){
pGain_A=20;
dGain_A=2;
iGain_A=5;
if (angle < 0){
	angle=angle*-1;
	if(!neg){
	errorI_A=0;
	}

	neg=1;
}
else{
if(neg){
errorI_A=0;
}
neg=0;
}

errorI_A = errorI_A + angle;
if(errorI_A>200){
errorI_A=200;
}
errorD_A = angle - prev_angle;

output_A = (pGain_A*angle) + (iGain_A*errorI_A) + (dGain_A*errorD_A);



if (output_A > 255){
	duty_A=255;
}
else if (output_A < 0){
	duty_A=0;
}
else{
	duty_A=output_A;
}



if(neg){
OCR0A=0;
OCR0B=duty_A;
}
else{
OCR0A=duty_A;
OCR0B=0;
}
prev_angle = angle;


///// Distance Control ///////

pGain_D=1.7;
dGain_D=1;
iGain_D=0.2;
error=(distance_setPoint-y)*-1;
output_D = error*pGain_D + (error-prev_dist)*dGain_D;// + errorI_D * iGain;

prev_dist = y;

if(output_D > 250){
duty_D = 250;
}

else if(output_D < 0) {
duty_D =0;
}
else {
 duty_D=output_D;
}

OCR2A=duty_D;

}
//////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////
void init(){
//set up timer for pulse emission 
TCCR3B=(1<<CS30)|(1<<CS31);
TIMSK3=(1<<OCIE3A);
OCR3A=30000;

//set up 40kHz PWM wave for pulse
//DDRD = (1<<PIND5);
TCCR1B = (1<<WGM12)|(1<<CS11);

TCCR1A = (0<<COM1A0) | (0<<COM1A1) |(0<<COM1B0) | (1<<COM1B1)| (0<<WGM10) | (0<<WGM11) ;

OCR1A = 41;
//OCR1B = 100 ; OCR1A is set to toggle, so OR1B does not do anything

//Set up external int
//set up on pin B2 & D3: INT2 & INT1
EIMSK=0;//INT2 & INT1 are disabled
EICRA=60;//Set external ints for rising edge detection

//Set up motor controller
	//set pwm pin to output
 	DDRB = (1<<PINB3) | (1<<PINB4) ;

 	
	TCCR0B = 1 ;  
	
 	TCCR0A = (0<<COM0A0) | (1<<COM0A1) |(0<<COM0B0) | (1<<COM0B1)| (1<<WGM00) | (1<<WGM01) ; 
    OCR0A = 0 ; // duty cycle is set by this.  duty cycle = OCROA/255
	OCR0B = 0 ;

	
	DDRD = (1<<PIND7) | (1<<PIND6)|(1<<PIND5);//|(1<<PIND1)|(1<<PIND0); 

	TCCR2B = 1;

	TCCR2A = (0<<COM2A0) | (1<<COM2A1) |(0<<COM2B0) | (1<<COM2B1)| (1<<WGM20) | (1<<WGM21) ; 
    OCR2A = 0 ; // duty cycle is set by this.  duty cycle = OCROA/255
	OCR2B = 0 ;



sei();
}
//////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////
void emitPulse(void){
pulse_start=TCNT3;
//EIMSK= 0;
TCCR1A = (1<<COM1A0);
while(TCNT3<(pulse_start+pulseLength));
TCCR1A = (0<<COM1A0);
}
////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////
void main(){
init();

while(1);

}
////////////////////////////////////////////////////////////////
